#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <linux/dvb/dmx.h>
#include <linux/dvb/frontend.h>

#include "getstream.h"

struct diseqc_cmd {
	struct dvb_diseqc_master_cmd cmd;
	uint32_t wait;
};

void diseqc_send_msg(int fd, fe_sec_voltage_t v, struct diseqc_cmd *cmd,
		fe_sec_tone_mode_t t, fe_sec_mini_cmd_t b) {

	ioctl(fd, FE_SET_TONE, SEC_TONE_OFF);
	ioctl(fd, FE_SET_VOLTAGE, v);
	usleep(15 * 1000);

	ioctl(fd, FE_DISEQC_SEND_MASTER_CMD, &cmd->cmd);
	usleep(cmd->wait * 1000);
	usleep(15 * 1000);

	ioctl(fd, FE_DISEQC_SEND_BURST, b);
	usleep(15 * 1000);

	ioctl(fd, FE_SET_TONE, t);
}

/*
 * DIgital Satellite Equipment Control,
 * Specification is available from http://www.eutelsat.com/
 */
static int head_diseqc(int secfd, int satno, int voltage, int tone) {
	struct diseqc_cmd cmd = { {{0xe0, 0x10, 0x38, 0xf0, 0x00, 0x00}, 4}, 0 };

	/* 
	 * param: high nibble: reset bits, low nibble set bits,
	 * bits are: option, position, polarizaion, band
	 */
	cmd.cmd.msg[3] = 0xf0 |
		(((satno * 4) & 0x0f) |
		(voltage == SEC_VOLTAGE_13 ? 1 : 0) |
		(tone == SEC_TONE_ON ? 0 : 2));

	diseqc_send_msg(secfd,
			voltage,
			&cmd,
			tone,
			(satno / 4) % 2 ? SEC_MINI_B : SEC_MINI_A);

	return 1;
}						     

/*
 * Dump FrontEnd Status byte as clear text returned
 * by GET_STATUS or GET_EVENT ioctl
 *
 */
char *fe_decode_status(int status) {
	static char	str[256];

	str[0]=0x0;

	if (status & FE_HAS_SIGNAL) 
		strcat(str, "HAS_SIGNAL ");
	if (status & FE_HAS_CARRIER)
		strcat(str, "HAS_CARRIER ");
	if (status & FE_HAS_VITERBI) 
		strcat(str, "HAS_VITERBI ");
	if (status & FE_HAS_SYNC) 
		strcat(str, "HAS_SYNC ");
	if (status & FE_HAS_LOCK) 
		strcat(str, "HAS_LOCK ");
	if (status & FE_TIMEDOUT) 
		strcat(str, "TIMEDOUT ");
	if (status & FE_REINIT) 
		strcat(str, "REINIT ");

	return str;
}

/*
 * Wait for the frontend to deliver an event
 * Return to callee if either TIMEDOUT or HAS_LOCK
 *
 */
int fe_get_event(int fd) {
	struct pollfd pfd;
	struct dvb_frontend_event event;
	int	res;

	pfd.fd = fd;
	pfd.events = POLLIN;

	do {
		if (!poll(&pfd, 1, 10000))
			continue;

		if (!(pfd.revents & POLLIN))
			continue;

		res=ioctl(fd, FE_GET_EVENT, &event);

		/*
		 * EOVERFLOW happens if we didnt poll for events
		 * lately and the event queue (8) overflowed inbetween.
		 *
		 */
		if (res < 0 && errno != EOVERFLOW) {
			perror("FE_GET_EVENT failed");
			exit(-1);
		}

		if (res >= 0 && event.status)
			dprintf("Status: %02x %s\n", event.status, fe_decode_status(event.status));

	} while (((event.status & FE_HAS_LOCK) == 0) && ((event.status & FE_TIMEDOUT) == 0));

	return(event.status);
}

static int fe_tune_dvbs(struct adapter_s *adapter) {
	struct dvb_frontend_parameters	feparams;
	int				voltage, tone, i;

	memset(&feparams, 0, sizeof(struct dvb_frontend_parameters));

	if (strcasecmp(adapter->dvbs.t_pol, "h") == 0) {
		voltage=SEC_VOLTAGE_18;
	} else if(strcasecmp(adapter->dvbs.t_pol, "v") == 0) {
		voltage=SEC_VOLTAGE_13;
	} else {
		perror("Unknown polarity in dvb-s transponder config");
		exit(-1);
	}

	if (ioctl(adapter->fefd, FE_SET_VOLTAGE, voltage) < 0) {
		fprintf(stderr, "ioctl FE_SET_VOLTAGE failed\n");
		exit(-1);
	}

	if (adapter->dvbs.t_freq > 2200000) {
		if (adapter->dvbs.t_freq < adapter->dvbs.lnb_slof) {
			feparams.frequency=(adapter->dvbs.t_freq-adapter->dvbs.lnb_lof1);
			tone = SEC_TONE_OFF;
		} else {
			feparams.frequency=(adapter->dvbs.t_freq-adapter->dvbs.lnb_lof2);
			tone = SEC_TONE_ON;
		}
	} else {
		feparams.frequency=adapter->dvbs.t_freq;
	}

	feparams.inversion=INVERSION_AUTO;
	feparams.u.qpsk.symbol_rate=adapter->dvbs.t_srate;
	feparams.u.qpsk.fec_inner=FEC_AUTO;

	if (adapter->dvbs.t_diseqc) {
		for(i=0;i<=1;i++) {
			if (debuglvl >= 1)
				fprintf(stderr, "fe: diseqc sending command %d\n", i);

			if (!head_diseqc(adapter->fefd, adapter->dvbs.t_diseqc-1, voltage, tone)) {
				fprintf(stderr, "fe: diseqc failed to send\n");
				exit(-1);
			}

			if (debuglvl >= 1) 
				fprintf(stderr, "fe: diseqc send successful\n");
			sleep(1);
		}
	} else {
		if (ioctl(adapter->fefd, FE_SET_VOLTAGE, voltage) < 0) {
			fprintf(stderr, "ioctl FE_SET_VOLTAGE failed\n");
			exit(-1);
		}

		if (ioctl(adapter->fefd, FE_SET_TONE, tone) < 0) {
			perror("ERROR setting tone\n");
		}
	}

	dprintf("FE_SET_FRONTEND:\n");
	dprintf("\ttone = %d\n", tone);
	dprintf("\tdiseqc = %d\n", adapter->dvbs.t_diseqc);
	dprintf("\tt->freq = %lu\n", adapter->dvbs.t_freq);
	dprintf("\tlnb->lof1 = %lu\n", adapter->dvbs.lnb_lof1);
	dprintf("\tlnb->lof2 = %lu\n", adapter->dvbs.lnb_lof2);
	dprintf("\tlnb->slof = %lu\n", adapter->dvbs.lnb_slof);
	dprintf("\tfeparams.frequency = %d\n", feparams.frequency);
	dprintf("\tfeparams.inversion = %d\n", feparams.inversion);
	dprintf("\tfeparams.u.qpsk.symbol_rate = %d\n", feparams.u.qpsk.symbol_rate);
	dprintf("\tfeparams.u.qpsk.fec_inner = %d\n", feparams.u.qpsk.fec_inner);

	if (ioctl(adapter->fefd, FE_SET_FRONTEND, &feparams) < 0) {
		fprintf(stderr, "ioctl FE_SET_FRONTEND failed\n");
		exit(-1);
	}

	return 0;
}

int fe_tune_dvbt(struct adapter_s *adapter) {
	struct dvb_frontend_parameters	feparams;

	memset(&feparams, 0, sizeof(struct dvb_frontend_parameters));

	feparams.inversion = INVERSION_AUTO;
	feparams.frequency = adapter->dvbt.freq;

	switch(adapter->dvbt.bandwidth) {
		case(0):
			feparams.u.ofdm.bandwidth=BANDWIDTH_AUTO;
			break;
		case(6):
			feparams.u.ofdm.bandwidth=BANDWIDTH_6_MHZ;
			break;
		case(7):
			feparams.u.ofdm.bandwidth=BANDWIDTH_7_MHZ;
			break;
		case(8):
			feparams.u.ofdm.bandwidth=BANDWIDTH_8_MHZ;
			break;
		default:
			fprintf(stderr, "fe: Unknown DVB-T bandwidth %d\n", adapter->dvbt.bandwidth);
			exit(-1);
	}

	feparams.u.ofdm.code_rate_HP = FEC_AUTO;
	feparams.u.ofdm.code_rate_LP = FEC_AUTO;

	switch(adapter->dvbt.modulation) {
		case(0):
			feparams.u.ofdm.constellation = QAM_AUTO;
			break;
		case(16):
			feparams.u.ofdm.constellation = QAM_16;
			break;
		case(32):
			feparams.u.ofdm.constellation = QAM_32;
			break;
		case(64):
			feparams.u.ofdm.constellation = QAM_64;
			break;
		case(128):
			feparams.u.ofdm.constellation = QAM_128;
			break;
		case(256):
			feparams.u.ofdm.constellation = QAM_256;
			break;
		default:
			fprintf(stderr, "fe: Unknown DVB-T modulation %d\n", adapter->dvbt.modulation);
			exit(-1);
	}

	switch(adapter->dvbt.tmode) {
		case(0):
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_AUTO;
			break;
		case(2):
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_2K;
			break;
		case(8):
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_8K;
			break;
		default:
			fprintf(stderr, "fe: Unknown DVB-T transmission mode %d\n", adapter->dvbt.tmode);
			exit(-1);
	}

	switch(adapter->dvbt.guard) {
		case(0):
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_AUTO;
			break;
		case(4):
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_4;
			break;
		case(8):
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_8;
			break;
		case(16):
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_16;
			break;
		case(32):
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_32;
			break;
		default:
			fprintf(stderr, "fe: Unknown DVB-T guard interval %d\n", adapter->dvbt.guard);
			exit(-1);
	}


	switch(adapter->dvbt.hierarchy) {
		case(-1):
			feparams.u.ofdm.hierarchy_information = HIERARCHY_NONE;
			break;
		case(0):
			feparams.u.ofdm.hierarchy_information = HIERARCHY_AUTO;
			break;
		case(1):
			feparams.u.ofdm.hierarchy_information = HIERARCHY_1;
			break;
		case(2):
			feparams.u.ofdm.hierarchy_information = HIERARCHY_2;
			break;
		case(4):
			feparams.u.ofdm.hierarchy_information = HIERARCHY_4;
			break;
		default:
			fprintf(stderr, "fe: Unknown DVB-T hierarchy %d\n", adapter->dvbt.hierarchy);
			exit(-1);
	}

	if (ioctl(adapter->fefd, FE_SET_FRONTEND, &feparams) < 0) {
		fprintf(stderr, "ioctl FE_SET_FRONTEND failed\n");
		exit(-1);
	}

	return 0;
}

static int fe_tune_dvbc(struct adapter_s *adapter) {
	struct dvb_frontend_parameters	feparams;

	memset(&feparams, 0, sizeof(struct dvb_frontend_parameters));

	feparams.frequency = adapter->dvbc.freq;
	feparams.inversion=INVERSION_AUTO;
	feparams.u.qam.symbol_rate = adapter->dvbc.srate;
	switch(adapter->dvbc.modulation) {
		case -1: feparams.u.qam.modulation = QPSK; break;
		case 0: feparams.u.qam.modulation = QAM_AUTO; break;
		case 16: feparams.u.qam.modulation = QAM_16; break;
		case 32: feparams.u.qam.modulation = QAM_32; break;
		case 64: feparams.u.qam.modulation = QAM_64; break;
		case 128: feparams.u.qam.modulation = QAM_128; break;
		case 256: feparams.u.qam.modulation = QAM_256; break;
		default:
			fprintf(stderr, "Unknown modulation %d\n", adapter->dvbc.modulation);
			exit(-1);
	}
	switch(adapter->dvbc.fec) {
		case 0: feparams.u.qam.fec_inner = FEC_NONE; break;
		case 1: feparams.u.qam.fec_inner = FEC_1_2; break;
		case 2: feparams.u.qam.fec_inner = FEC_2_3; break;
		case 3: feparams.u.qam.fec_inner = FEC_3_4; break;
		case 4: feparams.u.qam.fec_inner = FEC_4_5; break;
		case 5: feparams.u.qam.fec_inner = FEC_5_6; break;
		case 6: feparams.u.qam.fec_inner = FEC_6_7; break;
		case 7: feparams.u.qam.fec_inner = FEC_7_8; break;
		case 8: feparams.u.qam.fec_inner = FEC_8_9; break;
		case 9: feparams.u.qam.fec_inner = FEC_AUTO; break;
		default:
			fprintf(stderr, "Unknown fec %d\n", adapter->dvbc.fec);
			exit(-1);
	}

	if (ioctl(adapter->fefd, FE_SET_FRONTEND, &feparams) < 0) {
		fprintf(stderr, "ioctl FE_SET_FRONTEND failed\n");
		exit(-1);
	}

	return 0;
}

int fe_tune(struct adapter_s *adapter) {
	char				fename[128];
	int				status;


	sprintf(fename, "/dev/dvb/adapter%d/frontend0", adapter->no);

	adapter->fefd=open(fename, O_RDWR);

	if (!adapter->fefd) {
		perror("Error opening dvb frontend");
		exit(-1);
	}

	switch(adapter->type) {
		case (AT_DVBS):
			fe_tune_dvbs(adapter);
			break;
		case(AT_DVBT):
			fe_tune_dvbt(adapter);
			break;
		case(AT_DVBC):
			fe_tune_dvbc(adapter);
			break;
		default:
			printf("Unknown adapter type\n");
			exit(-1);
	}

	do {
		status=fe_get_event(adapter->fefd);
	} while (!(status && FE_HAS_LOCK) && !(status && FE_TIMEDOUT));

	/* TODO We should return the condition whether we suceeded */

	return 1;
}

